/*
 *   Copyright (C) 2012 - 2016 by Ivan Cukic <ivan.cukic@kde.org>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License or (at your option) version 3 or any later version
 *   accepted by the membership of KDE e.V. (or its successor approved
 *   by the membership of KDE e.V.), which shall act as a proxy
 *   defined in Section 14 of version 3 of the license.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef UTILS_MERGE_INTO_H
#define UTILS_MERGE_INTO_H

namespace kamd {
namespace utils {

template <typename Container>
inline void merge_into(Container &into, const Container &from)
{
    typename Container::iterator into_begin = into.begin();
    typename Container::iterator into_end = into.end();
    typename Container::const_iterator from_begin = from.begin();
    typename Container::const_iterator from_end = from.end();

    while (from_begin != from_end) {
        while (into_begin != into_end && *from_begin >= *into_begin)
            into_begin++;

        into_begin = into.insert(into_begin, *from_begin);
        into_begin++;
        from_begin++;
    }
}

} // namespace utils
} // namespace kamd

#endif // UTILS_MERGE_INTO_H
